# Copyright The Numerical Algorithms Group Limited 1992-94.
#
# This program translates oplist-<suffix>.tex files so that:
#   1) names of system functions and parameters appear in {\tt }
#   2) user-specifiable parameters appear in italics or math mode
#
# Specific transformations:
#   1)  \opdata{<title>}{2}{3}{4}{<doc>} forms:
#       a) <title> and <doc> get transformed as described below.
#       b) arguments 2,3,4 remain unchanged
#
#   2)  All \smath{} forms are transformed as described below.
#
#   3)  \xxdata{} forms are skipped over. They must be on a single line.
#       (RSS, 3/17/92: it is assumed that these are commented out)
#
# Syntax of <title>
#       <title>   :: <form> [\optand <form>]* [\opand <form>]
#          The \optand and \opand are used for vertically stacking titles.
#           The entire title is wrapped with {\it ..}.
#       <form>    :: <name> ( [<arglist>] [<optPart>] )
#          Every form can have at most one \opt{} part for optional arguments.
#       <arglist> :: <name> <argtail>
#          A form always has open/close prens but arguments are optional.
#       <argtail> :: [, <name>]*
#          Arguments are separated by commas
#       <optPart> :: \opt{ <optBody> }
#          The optional part always has a body
#       <optBody> :: <arglist> | <argtail>
#          There are two forms \opt{a} when the first argument is optional,
#          and \opt{,a,b} when a later argument is optional
# Translations:
#          <name>   -->  {\tt <name>}
#          ( , )    -->  {\tt (}    {\tt ,}    {\tt )}
#          \opt{x}  -->  \lbrack translationOf(x) \rbrack
#
# Translations of \smath{x}:
#   Entire \smath{x} is formatted in math mode with translations as follows:
#   If x has the form <name>(<arglist>) [\$ <single-letter>],
#   where <name> is alphabetic with 3 or more alphabetic characters,
#   and where <arglist> can have an \optinner{} term with similar
#   syntax to that of \opt{} in <title>.
#          <name>   --> {\tt <name>}
#          ( , )    --> {\tt (}    {\tt ,}    {\tt )}
#          All other \smath{x} forms are formatted as $x$.

#-----------------------------------------------------------------------------
#                        MAIN PROGRAM
#-----------------------------------------------------------------------------

BEGIN   {
        TRUE = 1
        FALSE = 0
        sep = "% ----------------------------------------------------------------------"
}

# skip over (exclude) comments
/^%/ {
        next
}

/^\\opdata/ {
        print sep
        printf "%s","\\opdata{"            # print \opdata{
        $0 = substr($0,9)                  # advance past \opdata{
        header()                           # print and process first argument
        k = index($0,"\}")
        $0 = substr($0,k+1)                # advance past }
        printf "}%%\n"                      # print matching }
        $0 =findSmath($0)
        next
}
/^\\keydata/ {
        $0 =findSmath($0)
        next
}
/^\\xxdata/ {                              # skip over \xxdata{} lines
        next
}
# otherwise
{
        $0 = findSmath($0)                  # transform \smath{} in other lines
        next
}
#-------------------------------------------------------------------------
#                        TRANSLATE <title>
#-------------------------------------------------------------------------
function header() {
        printf "{"                                 # wrap braces around whole argument
        $0 = headform($0)
        k = index($0,"\\optand")
        while (k != 0) {
           printf "%s",substr($0,1,k+6)
           $0 = substr($0,k+7)
           $0 = headform($0)
           k = index($0,"\\optand")
        }
        k = index($0,"\\opand")
        while (k != 0) {
           printf "%s",substr($0,1,k+5)
           $0 = substr($0,k+6)
           $0 = headform($0)
           k = index($0,"\\optand")
        }
        printf "}%%\n"                            # wrap closing brace
}
#-------------------------------------------------------------------------
#                        TRANSLATE <form>
#-------------------------------------------------------------------------
function headform(form,j,k) {
        form = dropLeadingBlanks(form)
        k = index(form,"(")
        opname = substr(form,1,k-1)            # name not including left pren
        printf "\\mbox{\\axiomFun{%s}}\\opLeftPren{}",opname
        form = substr(form,k+1)                # expose first argument
        form = dropLeadingBlanks(form)
        j = index(form,")")                    # check for empty argument
        if (j != 1) {
          c = getCommaIndex(form)              # try reading a comma
#  printf "%s%s%s%d%s","\n==>",form,"----------->",c,"\n"
          k = c
          if (c == 0) k = getDelimiterIndex(form)# if none, read a right pren
          firstArg = substr(form,1,k-1)
          printArg(firstArg)
          form = substr(form,k)                # expose comma after 1st arg
          form = commas(form)
        }
#-----------------------------------------------------------------------------
#                         TRANSLATE \opt{}
#--------------------------------------------------------------------------
        k = index(form,"\\opt{")
        k1= index(form,"\\opand")
        k2= index(form,"\\optand")
#   printf "%s%s%s%d%s%d%s%d%s","\n--->",form,"===>",k,"-->",k1,"-->",k2,"\n"
        if (k != 0 && (k1 == 0 || k < k1) && (k2 == 0 || k < k2)) {
           argBeforeBracket=substr(form,1,k-1)
           printArg(argBeforeBracket)
           printf "\\allowbreak $\\,\\lbrack$ "
           form = substr(form,k+5)
           k = closingBraceIndex(form)
           optform = substr(form,1,k-1)
           c = getCommaIndex(optform)              # try reading a comma
           if (c==0) {firstArg=optform} else {firstArg=substr(optform,1,c-1)}
           printArg(firstArg)
           optform=substr(optform,c)
           optform = commas(optform)
           form = optform substr(form,k+1)
           printf "$\\rbrack$"
        }
        k =index(form,"\)")                # advance past right pren
        form = substr(form,k+1)
        printf "\\opRightPren{}%%\n"
        return form
}

function printArg(arg) {
         if (arg != "")
           printf "{\\it %s}",arg
}

function commas(form,j,k) {                    # print out arguments
        while (substr(form,1,1) == ",") {
          form = substr(form,2)                # expose next argument
          k = getCommaIndex(form)
#  printf "%s%s%s%d%s","\n......",form,"----------->",k,"\n"
          # If no comma, go ONLY as far as a delimiter---remember you are
          # working the entire rest of the line here
          # Of course: stop earlier if you have an "\opt{}"
          rightPrenIndex=getDelimiterIndex(form)
#  printf "%s%d%s%s","\n>>>>>>  ",rightPrenIndex,form,"  <<<<\n"
          if (k == 0) {
            j = index(form,"\\opt{")
#   printf "%s%s%s%d%s%d%s","\n......",form,"-->",rightPrenIndex,"--------->",k,"\n"
          if (j != 0 && j < rightPrenIndex) {k=j} else {k = rightPrenIndex}
          }
#   printf "%s%s%s%d%s%d%s","\n......",form,"-->",rightPrenIndex,"--------->",k,"\n"
          if (k==0) {
            arg=form
            form= ""
          } else {
            arg=substr(form,1,k-1)
            form=substr(form,k)
          }
#         printf "%s","\{\\tt ,\}"
          printf ", \\allowbreak{}"
          printArg(arg)
        }
        return form
}
#-----------------------------------------------------------------------------
#               SEARCH FOR \smath{} WITHIN A STRING
#-----------------------------------------------------------------------------

function findSmath(line,k) {
        original=line
        k = index(line,"\\smath{")           # scan for \smath{}
        if (k == 0)
          return finish(original)            # if not found, print/return
        prefix=substr(line,1,k-1)
        line = substr(line,k+7)              # expose name(...) inside \smath{}
        k = closingBraceIndex(line)
        if (k==0)
          return finish(original)
        printf "%s\\smath{",prefix
        form = substr(line,1,k-1)            # get form inside \smath{}
        line = substr(line,k+1)
        spadTran(form)
        printf "}"                           # print closing brace for \smath
        return findSmath(line)
}

function finish(line) {                      # print out a line
          printf "%s%s",line,"\n"
          return
}

#-----------------------------------------------------------------------------
#                 Transform string inside \smath{}
#-----------------------------------------------------------------------------

function spadTran(form,k) {
#  printf "%s%s%s","\n========",form,"------\n"
        k = index(form,"(")
        form=dropTrailingBlanks(form)        # kill trailing blanks
        leftPrenIndex=index(form,"(")
        hasRightPren=0
        if (leftPrenIndex != 0) {
          afterPren=substr(form,leftPrenIndex+1)
          j=getDelimiterIndex(afterPren)
          if (j==length(afterPren)) {
            hasRightPren=1
          }
        }
        hasDollar  = (substr(form,length(form)-1,1) == "\$")
        if (hasDollar==1) {
          domain=substr(form,length(form),1)
          form=substr(form,1,length(form)-2)
          if (substr(form,length(form),1) == "\\") {     #kill backslash before $
            form=substr(form,1,length(form)-1)
          }
        }
        name = substr(form,1,k-1)
        m = length(name)
        looksOk = (hasRightPren || hasDollar)
        if (k == 0 || !looksOk) {
           if (k !=0 && isFunctionName(name)) {
             printf "\\mbox{\\bf %s}\\opLeftPren{}",name
             form = substr(form,k+1)      # expose first argument
           }
           printf "%s",form               # no "(" so quit
           return
        }
        if (!isFunctionName(name)) {
          printf "%s",form
          return
        }
        if (hasRightPren)                 # kill matching close pren
          form=substr(form,1,length(form)-1)

        printf "\\mbox{\\bf %s}\\opLeftPren{}",name
        form = substr(form,k+1)           # expose first argument
        k = getCommaIndex(form)           # is there a comma?
        if (k==0)
          k = getDelimiterIndex(form)     # if not, a delimiter?
        if (k==0) {
          printf "%s\\opRightPren{}",form
          if (hasDollar==1)
            printf "\\$%s",domain
          return
        }
        firstArg = substr(form,1,k-1)
        printf "%s",firstArg
        form = substr(form,k)
        form = spadCommas(form)

        #-------------------- \optinner handler --------------------
        #   here, form may look like "arg\optinner{, args}
        #-----------------------------------------------------------

        k = index(form,"\\optinner{")
        if (k != 0) {
           printf "%s\\bigLeftBrack ",substr(form,1,k-1)
           form = substr(form,k+10)
           k = closingBraceIndex(form)
           optform = substr(form,1,k-1)
           optform = spadCommas(optform)
           form = optform substr(form,k+1)
           printf "\\bigRightBrack "
         }
         #form should now be JUST a right pren
         k = index(form,"\)")
         form=substr(form,k+1)
         printf "%s\\opRightPren{}",form
         if (hasDollar==1)
           printf "\\$%s",domain
}

function spadCommas(form,k) {
        form = dropLeadingBlanks(form)
        while (substr(form,1,1) == ",") {
          form = substr(form,2)     # expose next argument
          k = getCommaIndex(form)   # get index of delimiter (comma/right pren)
          if (k==0) {
            k = getDelimiterIndex(form)
            j = index(form,"\\optinner")
            if (j != 0 && j < k) k=j
          }
          if (k==0) {
            arg = form
            form=""
          } else {
            arg = substr(form,1,k-1)
            form=substr(form,k)
          }
#         printf "{\\tt ,}%s",arg
          printf ", \\allowbreak{}%s",arg
        }
        return form
}
#-----------------------------------------------------------------------------
#                   SEARCH FUNCTIONS
#-----------------------------------------------------------------------------
function getCommaIndex(line,i, j) {    # Returns index of ","
        char = substr(line,1,1)
        count=1
        level=0
        while (char != ",") {
          if (char == "{")  {                     # move past {}
            while ((level > 1) || (char != "\}")) {
              if (char == "{")
                level++
              if (char == "}")
                level--
              count++
              line = substr(line,2)
              char = substr(line,1,1)
              if (length(line) == 0) {
                return 0
              }}
             level=0
            }
          else if (char == "(")   {             # move past ()
            while ((level > 1) || (char != ")")) {
              if (char == "(")
                level++
              if (char == ")")
                level--
              count++
              line = substr(line,2)
              char = substr(line,1,1)
              if (length(line) == 0) {
                return 0
              }}
            level=0
            }
          else if (char ~ /[\}\)\\]/) {
            return 0
          }
          count += 1
          line = substr(line,2)
          if (length(line)==0) return 0
          char = substr(line,1,1)
#printf "%s%d%s%s%s","\n---NEXT--->",count,"--->",line,"\n"
        }
        j = index(string,"\\op")
        if (j != 0 && j < count) count=j
#printf "%s%s%d%s","\n===>",string,count,"\n"
        return count
}
function closingBraceIndex(line,level,char,count) {
        char = substr(line,1,1)
        level=1
        count=1
        while ((level > 1) || (char != "}")) {
          if (char == "{")
            level++
          if (char == "}")
            level--
          line = substr(line,2)
          if (length(line) == 0) return 0
          char = substr(line,1,1)
          count++
        }
        return count
}

function getDelimiterIndex(line,i) {
#   getDelimiterIndex returns index of "{}()"
        char = substr(line,1,1)
        count=1
        level=0
        while (char !~ /[\}\)]/) {
#  printf "%s%s%s","\n--> ",char," <--\n"
          if (char == "{")  {                   # move past {}
            while ((level > 1) || (char != "}")) {
              if (char == "{")
                level++
              if (char == "}")
                level--
              count++
              line = substr(line,2)
              char = substr(line,1,1)
              if (length(line) == 0) {
                return 0
              }
            }
            level=0
          }
          else if (char == "(")   {             # move past ()
#           while ((level > 1) || (char != "}")) {
            while ((level > 1) || (char != ")")) {
              if (char == "(")
                level++
              if (char == ")")
                level--
              count++
              line = substr(line,2)
              char = substr(line,1,1)
              if (length(line) == 0) {
                return 0
              }}
            level=0
            }
          else if (char ~ /[\}\)]/) {
            return 0
          }
          count += 1
          line = substr(line,2)
          if (length(line)==0) return 0
          char = substr(line,1,1)

        }
        j = index(string,"\\op")
        if (j != 0 && j < count) count=j
        return count
}
#-----------------------------------------------------------------------------
#                   UTILITY FUNCTIONS
#-----------------------------------------------------------------------------

function isFunctionName(name) {
        if ((name == "D") || (name == "pi") || (name == "Ei") || (name == "e")|| (name == "hi") || (name == "lo"))
          return TRUE
        if ((length(name) < 3) || !isAlphabetic(name))
          return FALSE
        return TRUE
}

function isAlphabetic(name) {
         if (name ~ /^[A-Za-z][A-Za-z0-9\!?]*$/)
           return TRUE
         return FALSE
}

function min(m,n) {
        return m < n ? m : n
}

function dropLeadingBlanks(line) {
        while (substr(line,1,1) == " ") {
           printf " "
           line = substr(line,2)
        }
        return line
}

function dropTrailingBlanks(line) {
        while (substr(line,length(line),1) == " ") {
           line = substr(line,1,length(line)-1)
        }
        return line
}
