# Copyright The Numerical Algorithms Group Limited 1992-94.
#
# This awk script processes an .output file and creates a
# .tex file on standard output. One argument is expected
# and is of the form XXX.out

BEGIN           {
        stepNum = 0
        didOutput = 0
        didMsgs = 0
}

# kill nuisance messages
/Starts dribbling/      {
        next
}
/^  Warn    : Redefinition of macro/ {
        next
}
/\)spool/               {
        exit  # all done
}
/-- %/  {
        next
}

/^-- \\head/    {
         stepNum = 0
         print substr($0,4)
         print ""
         next
}

# handle htonly stuff (delete the lines)
/^-- \\htonly/   {
        while ($2 != "\\endhtonly")
                getline
        next
}

# handle discard stuff (delete the lines)
/^-- \\begin{discard}/   {
        while ($2 != "\\end{discard}")
                getline
        next
}

# delete lines between \begin{inputonly} and \end{inputonly}
/^-- \\begin{inputonly}/     {
        while ($2 != "\\end{inputonly}")
                getline
        next
}

# just drop texonly environments
/^-- \\begin{texonly}/ || /^-- \\end{texonly}/ || /^-- \\texonly/ || /^-- \\endtexonly/ {
        next
}

# handle piles
/^-- \\begin{spadsrc}/   {
        pileHandler()
        next
}

# handle two column examples
/^-- \\xtc\{/       {
        print "%"
        print "\\xtcLong{"
        next
}

# handle math mode
/^\\\[/ {
        mathHandler()
        next
}

/^%AXIOM STEP NUMBER:/  {
        stepNum = $4 + 0
        next
}

# kill )clear all messages
/^\)clear all/         {
        getline
        getline
        next
}

# kill blank lines
($1 == "")              {
        next
}

# process AXIOM comments (text)
/^--/                   {
        print substr($0,4)
        next
}

# handle type names
($1 == "Type:")         {
        typeHandler($0)
        next
}

# handle spad input. ASSUMPTION: there are not 2 consecutive
# spadcommands that do not produce any output.
(substr($0,1,1) != " ") {
        # translate special characters
        $0 = escapeAmpersand($0)
        gsub(/_/,"\\_")
        gsub(/{/,"\\{")
        gsub(/}/,"\\}")
        gsub(/#/,"\\#")
        gsub(/%/,"\\%")
        gsub(/\$/,"\\\$")
        len = length($0)
        if (substr($0,len,1) == " ")
                $0 = substr($0,1,len-1)
        printf "\\spadcommand{%s}\n}{\n",$0
        outputHandler()
        next
}


END             {
}

# user functions

function pileHandler(   i,j) {
        getline
        i = 0 + $2   # the number of lines
        print "\\vtop{"
        for (j = 1; j <= i; j++) {
          getline
          printVboxMessageLine($0)
        }
        print "}}{"
        outputHandler()
}

function typeHandler(line   ,i)      {
        if (((didOutput == 0) && (didMsgs == 0)) || ($1 == "Void")) {
                stepNum++
                print "\\vbox{\\hbox{\\quad}}%"
        }
        gsub(/%/,"\\%",line)
        gsub(/,/,", ",line)
        gsub(/->/,"\$\\rightarrow\$",line)
        numtoks = split(line,tok)
        len = 12
        if ($2 == "Void")
          printf "}{}{%%\n"
        else
          printf "}{%d}{%%\n",stepNum
        for (i = 2; i < numtoks; i++) {
          printf "%s ",tok[i]
          len += 1 + length(tok[i])
          if (len > 72) {
            printf "\n"
            len = 0
          }
        }
        printf "%s%%\n",tok[numtoks]
        didOutput = didMsgs = 0
}

function mathHandler() {
        if (didMsgs == 1)
                printf "\\vbox{%\n"
        print "\\\["
        while ($1 != "\\\]") {
                getline
                gsub(/#/,"\\#")
                gsub(/_/,"\\_")       # we use \sb for subscripts so this is ok
                if ($1 != "") {
                  if ((NF == 1) && ($1 == "true"))
                        print "{\\rm true}"
                  else if ((NF == 1) && ($1 == "false"))
                        print "{\\rm false}"
                  else
                        print
                }
        }
        if (didMsgs == 1)
                print "}% end of \\vbox"
        didOutput = 1
}

function finishXtcLong() {
        print "}{}{%"
        didOutput = didMsgs = 0
}

function outputHandler()        {
        # we might have some messages or other output that we want
        # to save in a verbatim environment. Next lines will be
        # one of
        #   . blank line
        #   . $$
        #   . start of text
        #   . start of msg
        #   . Type: ...
        #   . something else which we have to kluge

        # skip any blank lines
        getline
        while ($1 == "")
                getline
        # handle start of text
        if ($1 == "--")         {
                finishXtcLong()
                print substr($0,4)
                return
        }
        # handle start of math output
        if ($1 == "\\\[")        {
                mathHandler()
                return
        }
        # handle type printout
        if ($1 == "Type:")       {
                typeHandler($0)
                return
        }
        # handle output for "verbatim" mode
        print "\\vbox{\\Isize{}%"
        # continue until math mode or type or text
        while (($1 != "\\\[") && ($1 != "Type:") && ($1 != "--")) {
                if (substr($0,1,33) == "  Warn    : Redefinition of macro") {
                  getline
                  continue
                }
                if (substr($0,1,3) == "   ")
                  $0 = substr($0,4)
                printVboxMessageLine($0)
                # handle strange cases
                # 1. )set expose output
                if (match($0,/explicitly exposed/))
                        break
                if (match($0,/explicitly hidden/))
                        break
                # 2. output formatting
                if (match($0,/will be written to/))
                        break
                # 3. unknown function
                if (match($0,/Cannot find a definition for/)) {
                        # loop until blank line
                        getline
                        while ($1 != "") {
                                if (substr($0,1,3) == "   ")
                                  $0 = substr($0,4)
                                printVboxMessageLine($0)
                                getline
                        }
                        getline
                        break
                }
                getline
        }
        didMsgs = 1
        # handle start of math output
        if ($1 == "\\\[") {
                mathHandler()
                print "}% end of \\vbox" # end of \vbox around both math and output
                return
        }
        print "}% end of \\vbox" # end of \vbox around both math and output
        # handle type printout
        if ($1 == "Type:")
                typeHandler($0)
        # handle start of text
        else if ($1 == "--") {
                finishXtcLong()
                print substr($0,4)
        }
        else {
          finishXtcLong()
          printf("%% Possible Error: The line was: %s\n",$0)
        }

}

function printVboxMessageLine(line) {
         line = escapeAmpersand(line)
         gsub(/ /,"\\ ",line)
         gsub(/_/,"\\_",line)
         gsub(/%/,"\\%",line)
         gsub(/\$/,"\\$",line)
         gsub(/#/,"\\#",line)
         gsub(/{/,"\\{",line)
         gsub(/}/,"\\}",line)
         printf "\\vbox{\\outputMsg{} %s\\hfill}\n",line
}

function escapeAmpersand(line,   pa) {
# cannot get & and gsub to work together, escaped or not
        pa = index(line,"&")
        if (pa > 0)
          line = substr($0,1,pa-1)"\\&"escapeAmpersand(substr(line,pa+1))
        return line
}
